// SPI-basierte Kommunikation mit einem EEPROM-Speicher
#include "stm32g0xx_hal.h"

// Definitionen
// Der Speicherbaustein wird mithilfe von "Commands" angesprochen
// Sie legen die Art des Zugriffs fest: Lesen, Schreiben, Lesen des
// Statusregisters oder Freischalten des Schreibzugriffs
// Für eine bessere Lesbarkeit des Codes werden für die "Commands"
// die nachfolgenden Definitionen angelegt
#define CMD_25LCXX_WRITE 2 // Write
#define CMD_25LCXX_READ  3 // Read
#define CMD_25LCXX_WREN  6 // Write Enable
#define CMD_25LCXX_RDSR  5 // Read Status Register

// Definition für das "Write-In-Progress-Bit" des Statusregisters
#define SR_25LCXX_WIP    1 // Write In Progress

// Definition des GPIO-Anschlusses, welcher für das
// Slave-Select-Signal verwendet wird
#define NSS_GPIO GPIOA
#define NSS_GPIO_PIN GPIO_PIN_4

extern SPI_HandleTypeDef hspi1;

// Struct für den Datenaustausch mit dem Speicherbaustein
typedef struct {
    uint8_t  cmd;
    uint8_t  addrHighByte;
    uint8_t  addrLowByte;
    uint8_t  data[16];
} BUFFER_25LCXX;

volatile BUFFER_25LCXX txBuf, rxBuf;

// Funktionen zum Schreiben und Lesen des Slave-Select-Signals
void eepromSetNSS(GPIO_PinState val)
{
    HAL_GPIO_WritePin(NSS_GPIO, NSS_GPIO_PIN, val);
}

GPIO_PinState eepromGetNSS(void)
{
    return HAL_GPIO_ReadPin(NSS_GPIO, NSS_GPIO_PIN);
}

// Pollingbasierte Funktion für die Kommunikation mit dem Speicher
void eepromSPITransfer(BUFFER_25LCXX *txData,
                       BUFFER_25LCXX *rxData, uint16_t size)
{
    while (!eepromGetNSS()); // noch ein interruptbasierter Zugriff aktiv?
    eepromSetNSS(GPIO_PIN_RESET); // Slave-Select aktivieren
    // Pollingbasierten SPI-Transfer über HAL-Funktion ausführen
    HAL_SPI_TransmitReceive (&hspi1, (uint8_t*)txData,
                                     (uint8_t*)rxData, size, 10);
    eepromSetNSS(GPIO_PIN_SET); // Slave-Select deaktivieren
}

// Interruptbasierte Funktion für die Kommunikation mit dem Speicher
void eepromSPITransfer_IT(BUFFER_25LCXX *txData,
                          BUFFER_25LCXX *rxData, uint16_t size)
{
    while (!eepromGetNSS());// noch ein interruptbasierter Zugriff aktiv?
    eepromSetNSS(GPIO_PIN_RESET); // Slave-Select aktivieren
    // Interruptbasierten SPI-Transfer über HAL-Funktion ausführen
    HAL_SPI_TransmitReceive_IT (&hspi1, (uint8_t*)txData,
                                        (uint8_t*)rxData, size);
}

// Callback-Funktion, wird nach Ende des interruptb. Transfers aufgerufen
void HAL_SPI_TxRxCpltCallback(SPI_HandleTypeDef *hspi)
{
    if (hspi == &hspi1) {
        eepromSetNSS(GPIO_PIN_SET); // Slave-Select deaktivieren
    }
}


// EEPROM-Schreib- und Lesefunktionen
// (verwenden die oben definierten Grundfunktionen)

// Funktion zum Lesen des Statusregisters des Speicherbausteins
uint8_t eepromReadStatus(void)
{
    txBuf.cmd = CMD_25LCXX_RDSR;
    eepromSPITransfer (&txBuf,&rxBuf, 2);
    return rxBuf.addrHighByte;
}
// Funktion zum Schreiben von Daten
void eepromWriteData(uint16_t addr, uint8_t nData)
{
    uint8_t statusReg;
    do { // Write in Progress? (ggf. warten)
        statusReg = eepromReadStatus();
    } while(statusReg&SR_25LCXX_WIP);

    // Schreibfreigabe aktivieren (vor jedem Schreiben erforderlich)
    txBuf.cmd = CMD_25LCXX_WREN;   // Command: Write Enable
    eepromSPITransfer (&txBuf,&rxBuf, 1);

    // "Command" und Adresse in Struct setzen, dann Transfer starten
    txBuf.cmd = CMD_25LCXX_WRITE;
    txBuf.addrHighByte = addr>>8;
    txBuf.addrLowByte  = addr&0xFF;
    eepromSPITransfer_IT (&txBuf,&rxBuf, nData+3);
}

// Funktion zum Lesen von Daten
void eepromReadData(uint16_t addr, uint8_t nData)
{
    // "Command" und Adresse in Struct setzen, dann Transfer starten
    txBuf.cmd = CMD_25LCXX_READ;
    txBuf.addrHighByte = addr>>8;
    txBuf.addrLowByte  = addr&0xFF;
    eepromSPITransfer_IT (&txBuf,&rxBuf, nData+3);
}

// Beispielanwendung: SPI-EEPROM schreiben und lesen
// (verwendet die oben definierten Schreib- und Lesefunktionen)

void app() // wird in main() aufgerufen
{
    eepromSetNSS(1);
    // Beispieldaten an Adresse 0 ablegen
    txBuf.data[0]=11;
    txBuf.data[1]=12;
    eepromWriteData(0,2);

    // Beispieldaten an Adresse 0 wieder lesen
    // Durch die Schleife ist auch eine einfache Überprüfung
    // mit einem Logicanalyzer möglich
    while (1) {
        eepromReadData(0,2);
    }
}
